'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var react = require('react');
var jotai = require('jotai');

function useUpdateAtom(anAtom) {
  var writeOnlyAtom = react.useMemo(function () {
    return jotai.atom(null, function (_get, set, update) {
      return set(anAtom, update);
    });
  }, [anAtom]);
  return jotai.useAtom(writeOnlyAtom)[1];
}

var RESET = Symbol();
function atomWithReset(initialValue) {
  var anAtom = jotai.atom(initialValue, function (get, set, update) {
    if (update === RESET) {
      set(anAtom, initialValue);
    } else {
      set(anAtom, typeof update === 'function' ? update(get(anAtom)) : update);
    }
  });
  return anAtom;
}
function useResetAtom(anAtom) {
  var writeOnlyAtom = react.useMemo(function () {
    return jotai.atom(null, function (_get, set, _update) {
      return set(anAtom, RESET);
    });
  }, [anAtom]);
  return jotai.useAtom(writeOnlyAtom)[1];
}

function useReducerAtom(anAtom, reducer) {
  var _useAtom = jotai.useAtom(anAtom),
      state = _useAtom[0],
      setState = _useAtom[1];

  var dispatch = react.useCallback(function (action) {
    setState(function (prev) {
      return reducer(prev, action);
    });
  }, [setState, reducer]);
  return [state, dispatch];
}

function atomWithReducer(initialValue, reducer) {
  var anAtom = jotai.atom(initialValue, function (get, set, action) {
    return set(anAtom, reducer(get(anAtom), action));
  });
  return anAtom;
}

function atomFamily(initializeRead, initializeWrite, areEqual) {
  if (areEqual === void 0) {
    areEqual = Object.is;
  }

  var atoms = [];

  var createAtom = function createAtom(param) {
    var found = atoms.find(function (x) {
      return areEqual(x[0], param);
    });

    if (found) {
      return found[1];
    }

    var newAtom = jotai.atom(initializeRead(param), initializeWrite && initializeWrite(param));
    atoms.unshift([param, newAtom]);
    return newAtom;
  };

  createAtom.remove = function (p) {
    var index = atoms.findIndex(function (x) {
      return x[0] === p;
    });

    if (index >= 0) {
      atoms.splice(index, 1);
    }
  };

  return createAtom;
} // -----------------------------------------------------------------------

function useSelector(anAtom, selector, equalityFn) {
  if (equalityFn === void 0) {
    equalityFn = Object.is;
  }

  var sliceAtom = react.useMemo(function () {
    var prevSlice;
    var derivedAtom = jotai.atom(function (get) {
      var slice = selector(get(anAtom));

      if (equalityFn(prevSlice, slice)) {
        return prevSlice;
      }

      prevSlice = slice; // self contained mutation?

      return slice;
    });
    return derivedAtom;
  }, [anAtom, selector, equalityFn]);
  return jotai.useAtom(sliceAtom)[0];
}

function useAtomCallback(callback) {
  var anAtom = react.useMemo(function () {
    return jotai.atom(null, function (get, set, _ref) {
      var arg = _ref[0],
          resolve = _ref[1],
          reject = _ref[2];

      try {
        resolve(callback(get, set, arg));
      } catch (e) {
        reject(e);
      }
    });
  }, [callback]);

  var _useAtom = jotai.useAtom(anAtom),
      invoke = _useAtom[1];

  return react.useCallback(function (arg) {
    return new Promise(function (resolve, reject) {
      invoke([arg, resolve, reject]);
    });
  }, [invoke]);
}

exports.atomFamily = atomFamily;
exports.atomWithReducer = atomWithReducer;
exports.atomWithReset = atomWithReset;
exports.useAtomCallback = useAtomCallback;
exports.useReducerAtom = useReducerAtom;
exports.useResetAtom = useResetAtom;
exports.useSelector = useSelector;
exports.useUpdateAtom = useUpdateAtom;
